import os

import time
from dotenv import load_dotenv

if os.name == 'nt':
    os.system('color')

load_dotenv(os.getenv("WS2811_ENV_PATH", ".env"))

from ws2811_mqtt.args import init_args
init_args()

from ws2811_mqtt.logger import init_logger
init_logger()

from ws2811_mqtt.mqtt import init_mqtt
init_mqtt()

def main():
    while True:
        time.sleep(2)

if __name__ == "__main__":
    main()


import board
import neopixel
import os
import threading  # For running the alternating color loop
import time

from ws2811_mqtt.logger import log_client


NUM_LEDS = int(os.getenv("NUM_LEDS") or 50) # Number

# Initialize the NeoPixel strip
pixels = neopixel.NeoPixel(board.D18, NUM_LEDS, brightness=1, auto_write=True)
leds = [{"state": "OFF", "color": (255,255,255)} for _ in range(len(pixels))]
# pixels = [(0, 0, 0) for _ in range(NUM_LEDS)]

alternating_thread = None
alternating_colors_active = False
last_state = -1

alternating_colors_options = {
    "color_one": (255,255,0),
    "color_two": (0,255,255),
    "rate": 2,
    "transition": True,
}

def set_ac_option(key, value):
    global alternating_colors_options
    global alternating_colors_active

    log_client.info(f"[LEDS][%15s] set_ac_option called with key={key}, value={value}", "set_ac_option")
    alternating_colors_options[key] = value
    if alternating_colors_active:
        stop_alternating_colors()
        start_alternating_colors()


def manage_alternating_colors():
    global alternating_colors_active
    global alternating_colors_options
    global last_state
    try:
        state = -last_state
        while alternating_colors_active:
            state = -state
            for i in range(NUM_LEDS):
                set_l_on(i, alternating_colors_options.get("color_one") if state == 1 else alternating_colors_options.get("color_two"))
                if alternating_colors_options.get("transition"):
                    time.sleep(alternating_colors_options.get("rate") / NUM_LEDS)
                if not alternating_colors_active:
                    break
            log_client.info(f"[LEDS][%15s] state => {state}", "manage_alternating_colors")
            if not alternating_colors_options.get("transition"):
                time.sleep(alternating_colors_options.get("rate"))
    except Exception as e:
        log_client.error(f"[LEDS][%15s] Error in alternating colors: {e}", "manage_alternating_colors")

def start_alternating_colors():
    global alternating_thread, alternating_colors_active, alternating_colors_options
    if alternating_thread is not None and alternating_thread.is_alive():
        stop_alternating_colors()
        alternating_thread.join()  # Ensure the previous thread ends before starting a new one
    alternating_colors_active = True
    alternating_thread = threading.Thread(target=manage_alternating_colors, args=())
    alternating_thread.start()

def stop_alternating_colors():
    global alternating_colors_active
    alternating_colors_active = False
    if alternating_thread is not None:
        alternating_thread.join()

# Function to apply changes from the leds array to the pixels array
def set_led(led_index):
    try:
        if leds[led_index]["state"] == "OFF":
            pixels[led_index] = (0, 0, 0)
        else:
            pixels[led_index] = leds[led_index]["color"]
            log_client.debug(f"[LEDS][%15s] {led_index} => {leds[led_index]['state']}", "set_led")
    except Exception as e:
        log_client.error(f"[LEDS][%15s] Error applying LED changest to led {led_index}: {e}", "set_led")


# Function to check if an LED is on by verifying its color is not black (0, 0, 0)
def led_is_on(led_index):
    log_client.debug(f"[LEDS][%15s] LED index {led_index}", "led_is_on")
    log_client.debug(f"[LEDS][%15s] LED value {pixels[led_index]}", "led_is_on")
    led_on = leds[led_index]["state"] == "ON"
    return led_on

# Function to set a LED's color to black (0, 0, 0), effectively turning it off
def set_l_off(led_index):
    try:
        log_client.debug(f"[LEDS][%15s] LED value before {pixels[led_index]}", "set_l_off")
        pixels[led_index] = (0, 0, 0)
        log_client.debug(f"[LEDS][%15s] LED {led_index} color set to black.", "set_l_off")
        log_client.debug(f"[LEDS][%15s] LED value after  {pixels[led_index]}", "set_l_off")
    except Exception as e:
        log_client.error(f"[LEDS][%15s] Error setting LED color: {e}", "set_l_off")

# Function to set a LED's color to a specified value, defaulting to white (255, 255, 255)
def set_l_on(led_index, color=None):
    try:
        leds[led_index].update({"state": "ON", "color": color or leds[led_index]["color"]})
        set_led(led_index)
        log_client.debug(f"[LEDS][%15s] LED {led_index} color set to {color}.", "set_l_on")
    except Exception as e:
        log_client.error(f"[LEDS][%15s] Error setting LED color: {e}", "set_l_on")



from ws2811_mqtt.logger import log_client

def colr_tuple_to_st(color_tuple):
    try:
        # Check if the color tuple has exactly 3 components
        if len(color_tuple) != 3:
            raise ValueError("Color tuple must have three components")

        # Swap the first two elements back
        original_color = (color_tuple[1], color_tuple[0], color_tuple[2])

        # Convert the tuple into a comma-separated string
        color_string = ','.join(map(str, original_color))
        return color_string
    except Exception as e:
        log_client.error(f"[MQTT][%15s] Error creating color string from {color_tuple}: {e}", "colr_tuple_to_st")
        return None


def colr_str_to_tuple(color_string):
    try:
        # Parse the color string into a tuple of integers
        color = tuple(map(int, color_string.split(',')))

        # Check if the color tuple has exactly 3 components
        if len(color) != 3:
            raise ValueError("Color string must have three components")

        # Swap the first two elements
        swapped_color = (color[1], color[0], color[2])
        return swapped_color
    except Exception as e:
        log_client.error(f"[MQTT][%15s] Error parsing color string {color_string}: {e}", "colr_str_to_tuple")
        return None



import logging
from ws2811_mqtt.args import args_client

class CustomFormatter(logging.Formatter):

    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(levelname)-5s:%(message)s"

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

def init_logger():
    log_client = logging.getLogger("ws2811-mqtt")
    log_client.handlers = []  # Clear existing handlers
    log_client.propagate = False
    verbosity_levels = {
        0: logging.ERROR,
        1: logging.WARNING,
        2: logging.INFO,
        3: logging.DEBUG
    }
    log_client.setLevel(verbosity_levels.get(args_client.verbosity, logging.INFO))

    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(verbosity_levels.get(args_client.verbosity, logging.INFO))
    ch.setFormatter(CustomFormatter())
    log_client.addHandler(ch)
    return log_client

log_client = init_logger()


import argparse
args_client = None


def init_args():
    global args_client
    parser = argparse.ArgumentParser(description="Control the LED strip with MQTT and API")
    parser.add_argument('-v', '--verbosity', action='count', default=0, help="increase output verbosity, each v adds a verbosity level")
    args_client = parser.parse_args()


import os
import json
import paho.mqtt.client as mqtt
import yaml
import pprint
import logging

from ws2811_mqtt.logger import log_client
from ws2811_mqtt.leds import leds, set_led, NUM_LEDS, start_alternating_colors, stop_alternating_colors, set_ac_option
from ws2811_mqtt.utils import colr_str_to_tuple, colr_tuple_to_st



mqtt_client = None
device_config = None  # Declare a global variable for the device configuration

# MQTT settings
MQTT_BROKER = os.getenv("MQTT_BROKER")
if not MQTT_BROKER:
    raise ValueError("Missing MQTT_BROKER env var")
MQTT_PORT = os.getenv("MQTT_PORT")
if not MQTT_PORT:
    raise ValueError("Missing MQTT_PORT env var")
try:
    MQTT_PORT = int(MQTT_PORT)
except ValueError:
    raise ValueError("MQTT_PORT must be an integer")
MQTT_USER = os.getenv("MQTT_USER")
if not MQTT_USER:
    raise ValueError("Missing MQTT_USER env var")
MQTT_PASS = os.getenv("MQTT_PASS")
if not MQTT_PASS:
    raise ValueError("Missing MQTT_PASS env var")

MQTT_UID = os.getenv("MQTT_UID") or "ws2811-mqtt"
MQTT_TOPIC_STATUS = os.getenv("MQTT_TOPIC_STATUS") or f"homeassistant/device/{MQTT_UID}/config"


def publish_all_led(all_led_state):
    try:
        led_status = all_led_state["state"]
        topic = device_config.get("cmps").get(f"all_leds").get("state_topic")
        rgb_topic = device_config.get("cmps").get(f"all_leds").get("rgb_stat_t")
        mqtt_client.publish(f"{topic}", led_status.encode('utf-8'), retain=True)
        if "color" in all_led_state:
            led_status = colr_tuple_to_st(all_led_state["color"])
            mqtt_client.publish(f"{rgb_topic}", led_status.encode('utf-8'), retain=True)
            log_client.debug(f"[MQTT][%15s] Status: {all_led_state}", "publish_all_led")
        else:
            log_client.debug(f"[MQTT][%15s] Status: {all_led_state['state']}", "publish_all_led")
    except Exception as e:
        log_client.error(f"[MQTT][%15s] Error publishing LED status: {e}", "publish_all_led")


def publish_led(led_index):
    try:
        led = leds[led_index]
        led_status = leds[led_index]["state"]
        topic = device_config.get("cmps").get(f"led_{led_index + 1}").get("state_topic")
        rgb_topic = device_config.get("cmps").get(f"led_{led_index + 1}").get("rgb_stat_t")
        mqtt_client.publish(f"{topic}", led_status.encode('utf-8'), retain=True)
        led_status = colr_tuple_to_st(led["color"])
        mqtt_client.publish(f"{rgb_topic}", led_status.encode('utf-8'), retain=True)
        log_client.debug(f"[MQTT][%15s] Published LED status : Led #{led_index + 1} == {led['color']}", "publish_led")
    except Exception as e:
        log_client.error(f"[MQTT][%15s] Error publishing LED status: {e}", "publish_led")

def on_connect(client, userdata, flags, rc):
    log_client.info(f"[MQTT][%15s] Connected to broker", "on_connect")
    # Loop through LEDs and subscribe to their command topics
    for _, led_config in device_config['cmps'].items():
        for topic_type in ["command_topic", "rgb_cmd_t"]:
            topic = led_config.get(topic_type, False)
            if topic:
                client.subscribe(topic)
                log_client.info(f"[MQTT][%15s] Subscribed to {topic}", "on_connect")
            # rgb_command_topic = led_config['rgb_cmd_t']
            # client.subscribe(rgb_command_topic)
            # log_client.info(f"[MQTT][%15s] Subscribed to {rgb_command_topic}", "on_connect")

    # Publish initial status
    client.publish(MQTT_TOPIC_STATUS, json.dumps(device_config), retain=True)


def on_message(client, userdata, msg):
    try:
        payload = msg.payload.decode('utf-8')
        log_client.info(f"[MQTT][%15s] topic  : \"{msg.topic}\"", "on_message")
        log_client.info(f"[MQTT][%15s] payload: \"{msg.payload}\"", "on_message")
        if msg.topic == f"cmnd/{MQTT_UID}/all_leds/alternate":
            if payload == "ON":
                mqtt_client.publish(f"stat/{MQTT_UID}/all_leds", "OFF".encode('utf-8'), retain=True)
                start_alternating_colors()
            else:
                stop_alternating_colors()
            mqtt_client.publish(f"stat/{MQTT_UID}/all_leds/alternate", payload, retain=True)
        
        elif msg.topic == f"cmnd/{MQTT_UID}/all_leds/alternate/color_one/rgb":
            color_one = colr_str_to_tuple(payload)
            set_ac_option("color_one", color_one)
            mqtt_client.publish(f"stat/{MQTT_UID}/all_leds/alternate/color_one/rgb", payload.encode('utf-8'), retain=True)
        elif msg.topic == f"cmnd/{MQTT_UID}/all_leds/alternate/color_one":
            mqtt_client.publish(f"stat/{MQTT_UID}/all_leds/alternate/color_one", "ON".encode('utf-8'), retain=True)
        elif msg.topic == f"cmnd/{MQTT_UID}/all_leds/alternate/color_two/rgb":
            color_two = colr_str_to_tuple(payload)
            set_ac_option("color_two", color_two)
            mqtt_client.publish(f"stat/{MQTT_UID}/all_leds/alternate/color_two/rgb", payload.encode('utf-8'), retain=True)
        elif msg.topic == f"cmnd/{MQTT_UID}/all_leds/alternate/color_two":
            mqtt_client.publish(f"stat/{MQTT_UID}/all_leds/alternate/color_two", "ON".encode('utf-8'), retain=True)
        elif msg.topic == f"cmnd/{MQTT_UID}/all_leds/alternate/rate":
            rate = payload
            set_ac_option("rate", float(rate))
            mqtt_client.publish(f"stat/{MQTT_UID}/all_leds/alternate/rate", payload, retain=True)
        elif msg.topic == f"cmnd/{MQTT_UID}/all_leds/alternate/transition":
            transition = True if payload == "ON" else False
            set_ac_option("transition", transition)
            mqtt_client.publish(f"stat/{MQTT_UID}/all_leds/alternate/transition", payload, retain=True)
        elif msg.topic == f"cmnd/{MQTT_UID}/all_leds/rgb":
            stop_alternating_colors()
            for led_index in range(len(leds)):
                leds[led_index] = {"state": "ON", "color": colr_str_to_tuple(payload) }
                set_led(led_index)
                # publish_led(led_index)
                publish_all_led({"state": "ON", "color": colr_str_to_tuple(payload) })
            mqtt_client.publish(f"stat/{MQTT_UID}/all_leds/alternate", "OFF".encode('utf-8'), retain=True)
        elif msg.topic == f"cmnd/{MQTT_UID}/all_leds":
            stop_alternating_colors()
            for led_index in range(len(leds)):
                leds[led_index]["state"] = payload
                set_led(led_index)
                # publish_led(led_index)
                publish_all_led({"state": payload })
            mqtt_client.publish(f"stat/{MQTT_UID}/all_leds/alternate", "OFF".encode('utf-8'), retain=True)
        elif msg.topic in [led_config.get('command_topic') for led_config in device_config['cmps'].values()]:
            led_index = int(msg.topic.split('_')[-1]) -1
            leds[led_index]["state"] = payload
            set_led(led_index)
        elif msg.topic in [led_config.get('rgb_cmd_t') for led_config in device_config['cmps'].values()]:
            led_index = int(msg.topic.split(MQTT_UID)[-1].split('/')[1].split('_')[-1]) -1
            leds[led_index] = {"state": "ON", "color": colr_str_to_tuple(payload) }
            set_led(led_index)
        # publish_led(led_index)
    except Exception as e:
        log_client.error(f"[MQTT][%15s] Error processing message: {e}", "on_message")
        log_client.debug(f"[MQTT][%15s] Message payload : {msg.payload}", "on_message")
        log_client.debug(f"[MQTT][%15s] Message topic : {msg.topic}", "on_message")

def gen_leds_conf():
    # led_config = {}
    # Iterate over each component configuration in the device configuration
    # and update string values by replacing 'MQTT_UID' with its actual value.
    for _, component_values in device_config["cmps"].items():
        for key, value in component_values.items():
            if isinstance(value, str):
                component_values[key] = value.replace("MQTT_UID", MQTT_UID)

    # This loop generates a configuration dictionary for each LED by iterating over
    # a range from 1 to the total number of LEDs (`NUM_LEDS`), inclusive.
    # For each LED, it creates a unique identifier and sets its configuration,
    # including topics for state and command, which will be used with the MQTT broker,
    # then updates the device configuration with this information.
    # for i in range(1, NUM_LEDS + 1):
    #     led_id = f"led_{i}"
    #     led_number = f"{i:02}"  # Format as two digits
    #     led_config[led_id] = {
    #         "p": "light",
    #         "clrm": "rgb",
    #         "ret": True,
    #         "unique_id": f"ws2811_led_{led_number}",
    #         "rgb_stat_t": f"stat/{MQTT_UID}/led_{led_number}/rgb",
    #         "rgb_cmd_t": f"cmnd/{MQTT_UID}/led_{led_number}/rgb",
    #         "name": f"Led {led_number}",
    #         "command_topic": f"cmnd/{MQTT_UID}/led_{led_number}",
    #         "state_topic": f"stat/{MQTT_UID}/led_{led_number}"
    #     }

    # device_config["cmps"].update(led_config)
    # log_client.info("[MQTT][%15s] `LED configuration generated` :", "gen_leds_conf")
    if log_client.getEffectiveLevel() == logging.INFO:
        pprint.pprint(device_config, indent=2)


def init_mqtt():
    global mqtt_client, device_config  # Declare global to modify the external variables

    # Load device configuration once during initialization
    with open('ws2811_mqtt/device_config.yaml', 'r') as file:
        device_config = yaml.safe_load(file)
        # Replace placeholder MQTT_UID, and generate as many "Lights" as there are leds in HA
        gen_leds_conf()
    mqtt_client = mqtt.Client()
    mqtt_client.username_pw_set(username=MQTT_USER, password=MQTT_PASS)
    mqtt_client.on_connect = on_connect
    mqtt_client.on_message = on_message
    mqtt_client.connect(MQTT_BROKER, MQTT_PORT, 60)
    mqtt_client.loop_start()
    # for i in range(NUM_LEDS):
    #     publish_led(i)
